import unittest
from unittest import TestCase

from wheatley.row_generation import PlaceNotationGenerator
from . import gen_rows


class PlaceNotationGeneratorTests(TestCase):

    def test_plain_bob_minimus(self):
        stage = 4

        generator = PlaceNotationGenerator(stage, "&x1x1,2")

        rows = gen_rows(generator, 8)
        self.assertEqual([[2, 1, 4, 3],
                          [2, 4, 1, 3],
                          [4, 2, 3, 1],
                          [4, 3, 2, 1],
                          [3, 4, 1, 2],
                          [3, 1, 4, 2],
                          [1, 3, 2, 4],
                          [1, 3, 4, 2]],
                         rows)

    def test_plain_bob_minimus_bob(self):
        stage = 4

        generator = PlaceNotationGenerator(stage, "&x1x1,2")

        rows = gen_rows(generator, 6)
        generator.set_bob()
        for row in gen_rows(generator, 2):
            rows.append(row)

        self.assertEqual([[2, 1, 4, 3],
                          [2, 4, 1, 3],
                          [4, 2, 3, 1],
                          [4, 3, 2, 1],
                          [3, 4, 1, 2],
                          [3, 1, 4, 2],
                          [1, 3, 2, 4],
                          [1, 2, 3, 4]],
                         rows)

    def test_plain_bob_minimus_single(self):
        stage = 4

        generator = PlaceNotationGenerator(stage, "&x1x1,2")

        rows = gen_rows(generator, 6)
        generator.set_single()
        for row in gen_rows(generator, 2):
            rows.append(row)

        self.assertEqual([[2, 1, 4, 3],
                          [2, 4, 1, 3],
                          [4, 2, 3, 1],
                          [4, 3, 2, 1],
                          [3, 4, 1, 2],
                          [3, 1, 4, 2],
                          [1, 3, 2, 4],
                          [1, 3, 2, 4]],
                         rows)

    def test_plain_bob_doubles_plain(self):
        stage = 5

        generator = PlaceNotationGenerator(stage, "&5.1.5.1.5,2")

        rows = gen_rows(generator, 10)

        self.assertEqual([[2, 1, 4, 3, 5],
                          [2, 4, 1, 5, 3],
                          [4, 2, 5, 1, 3],
                          [4, 5, 2, 3, 1],
                          [5, 4, 3, 2, 1],
                          [5, 3, 4, 1, 2],
                          [3, 5, 1, 4, 2],
                          [3, 1, 5, 2, 4],
                          [1, 3, 2, 5, 4],
                          [1, 3, 5, 2, 4]],
                         rows)

    def test_plain_bob_doubles_two_bobs(self):
        stage = 5

        generator = PlaceNotationGenerator(stage, "&5.1.5.1.5,2")

        gen_rows(generator, 8)
        generator.set_bob()
        first_lead_end = gen_rows(generator, 2)
        # 4ths made at bob
        self.assertEqual([[1, 3, 2, 5, 4], [1, 2, 3, 5, 4]], first_lead_end)

        gen_rows(generator, 8)
        generator.set_bob()
        second_lead_end = gen_rows(generator, 2)
        # 4ths made at bob
        self.assertEqual([[1, 3, 2, 4, 5], [1, 2, 3, 4, 5]], second_lead_end)

    def test_resets_call_after_bob(self):
        stage = 5

        generator = PlaceNotationGenerator(stage, "&5.1.5.1.5,2")

        gen_rows(generator, 8)
        generator.set_bob()
        first_lead_end = gen_rows(generator, 2)
        # 4ths made at bob
        self.assertEqual([[1, 3, 2, 5, 4], [1, 2, 3, 5, 4]], first_lead_end)

        gen_rows(generator, 8)
        second_lead_end = gen_rows(generator, 2)
        # 2nds made at plain lead
        self.assertEqual([[1, 3, 2, 4, 5], [1, 3, 4, 2, 5]], second_lead_end)

    def test_stedman_doubles_plain(self):
        generator = PlaceNotationGenerator.stedman_doubles()

        rows = gen_rows(generator, 12)

        self.assertEqual([[2, 1, 3, 5, 4],
                          [2, 3, 1, 4, 5],
                          [3, 2, 4, 1, 5],
                          [2, 3, 4, 5, 1],
                          [2, 4, 3, 1, 5],
                          [4, 2, 3, 5, 1],
                          [4, 3, 2, 1, 5],
                          [3, 4, 2, 5, 1],
                          [4, 3, 5, 2, 1],
                          [4, 5, 3, 1, 2],
                          [5, 4, 3, 2, 1],
                          [5, 3, 4, 1, 2]],
                         rows)

    def test_stedman_doubles_first_single(self):
        generator = PlaceNotationGenerator.stedman_doubles()

        initial_rows = gen_rows(generator, 4)
        generator.set_single()
        single_rows = gen_rows(generator, 2)
        after_rows = gen_rows(generator, 6)

        self.assertEqual([[2, 1, 3, 5, 4],
                          [2, 3, 1, 4, 5],
                          [3, 2, 4, 1, 5],
                          [2, 3, 4, 5, 1]],
                         initial_rows)
        self.assertEqual([[2, 4, 3, 1, 5],
                          [4, 2, 3, 1, 5]],
                         single_rows)
        self.assertEqual([[4, 3, 2, 5, 1],
                          [3, 4, 2, 1, 5],
                          [4, 3, 1, 2, 5],
                          [4, 1, 3, 5, 2],
                          [1, 4, 3, 2, 5],
                          [1, 3, 4, 5, 2]],
                         after_rows)

    def test_stedman_doubles_second_single(self):
        generator = PlaceNotationGenerator.stedman_doubles()

        gen_rows(generator, 10)
        generator.set_single()
        single_rows = gen_rows(generator, 2)

        self.assertEqual([[5, 4, 3, 2, 1],
                          [5, 3, 4, 2, 1]],
                         single_rows)


    def test_grandsire_triples_bob(self):
        generator = PlaceNotationGenerator.grandsire(7)

        gen_rows(generator, 10)
        generator.set_bob()
        bob_rows = gen_rows(generator, 5)

        self.assertEqual([[5, 7, 1, 6, 2, 4, 3],
                          [5, 1, 7, 2, 6, 3, 4],
                          [1, 5, 7, 6, 2, 4, 3],
                          [1, 7, 5, 2, 6, 3, 4],
                          [7, 1, 5, 6, 2, 4, 3]],
                         bob_rows)

    def test_grandsire_triples_single(self):
        generator = PlaceNotationGenerator.grandsire(7)

        gen_rows(generator, 10)
        generator.set_single()
        single_rows = gen_rows(generator, 5)

        self.assertEqual([[5, 7, 1, 6, 2, 4, 3],
                          [5, 1, 7, 2, 6, 3, 4],
                          [1, 5, 7, 6, 2, 4, 3],
                          [1, 5, 7, 2, 6, 3, 4],
                          [5, 1, 7, 6, 2, 4, 3]],
                         single_rows)


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import TestCase

from wheatley.row_generation import PlainHuntGenerator
from . import gen_rows


class PlainHuntGeneratorTests(TestCase):

    def test_reset_handstroke(self):
        stage = 3
        generator = PlainHuntGenerator(stage)

        initial_rows = gen_rows(generator, 4)
        self.assertEqual([[2, 1, 3],
                          [2, 3, 1],
                          [3, 2, 1],
                          [3, 1, 2]],
                         initial_rows)
        generator.reset()
        # Starts again after rounds
        second_rows = gen_rows(generator, 4)
        self.assertEqual([[2, 1, 3],
                          [2, 3, 1],
                          [3, 2, 1],
                          [3, 1, 2]],
                         second_rows)

    def test_singles(self):
        stage = 3
        generator = PlainHuntGenerator(stage)
        rows = gen_rows(generator, 6)
        self.assertEqual([[2, 1, 3],
                          [2, 3, 1],
                          [3, 2, 1],
                          [3, 1, 2],
                          [1, 3, 2],
                          [1, 2, 3]],
                         rows)

    def test_minimus(self):
        stage = 4
        generator = PlainHuntGenerator(stage)

        rows = gen_rows(generator, 8)
        self.assertEqual([[2, 1, 4, 3],
                          [2, 4, 1, 3],
                          [4, 2, 3, 1],
                          [4, 3, 2, 1],
                          [3, 4, 1, 2],
                          [3, 1, 4, 2],
                          [1, 3, 2, 4],
                          [1, 2, 3, 4]],
                         rows)

    def test_cinques(self):
        stage = 11
        generator = PlainHuntGenerator(stage)

        rows = gen_rows(generator, 3)
        self.assertEqual([[2, 1, 4, 3, 6, 5, 8, 7, 10, 9, 11],
                          [2, 4, 1, 6, 3, 8, 5, 10, 7, 11, 9],
                          [4, 2, 6, 1, 8, 3, 10, 5, 11, 7, 9]],
                         rows)

    def test_maximus(self):
        stage = 12
        generator = PlainHuntGenerator(stage)

        rows = gen_rows(generator, 3)
        self.assertEqual([[2, 1, 4, 3, 6, 5, 8, 7, 10, 9, 12, 11],
                          [2, 4, 1, 6, 3, 8, 5, 10, 7, 12, 9, 11],
                          [4, 2, 6, 1, 8, 3, 10, 5, 12, 7, 11, 9]],
                         rows)

    def test_minimus_custom_start_row(self):
        stage = 4
        generator = PlainHuntGenerator(stage, "4321")

        rows = gen_rows(generator, 8)
        self.assertEqual([[3, 4, 1, 2],
                          [3, 1, 4, 2],
                          [1, 3, 2, 4],
                          [1, 2, 3, 4],
                          [2, 1, 4, 3],
                          [2, 4, 1, 3],
                          [4, 2, 3, 1],
                          [4, 3, 2, 1]],
                         rows)

if __name__ == '__main__':
    unittest.main()



"""
A module to contain a Bell class used to encapsulate the idea of a Bell and remove off-by-one
errors related to deciding whether the treble is bell #0 or bell #1.
"""

from typing import Any


BELL_NAMES = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "0", "E", "T", "A", "B", "C", "D"]
MAX_BELL = len(BELL_NAMES)


class Bell:
    """ A class to encapsulate the idea of a bell. """

    @classmethod
    def from_str(cls, bell_str: str) -> 'Bell':
        """
        Generates a Bell object from a string representing that bell's name.
        This works according to the standard convention, so Bell.from_str('1') will represent
        the treble, and Bell.from_str('T') will represent the twelfth.
        """
        try:
            index = BELL_NAMES.index(bell_str)
        except ValueError as e:
            raise ValueError(f"'{bell_str}' is not known bell symbol") from e

        return cls(index)

    @classmethod
    def from_number(cls, bell_num: int) -> 'Bell':
        """
        Generates a Bell from a 1-indexed number, so Bell.from_number(1) will return a Bell
        representing the treble.
        """
        return cls(bell_num - 1)

    @classmethod
    def from_index(cls, bell_index: int) -> 'Bell':
        """
        Generates a Bell from a 0-indexed number, so Bell.from_number(0) will return a Bell
        representing the treble.
        """
        return cls(bell_index)

    def __init__(self, index: int) -> None:
        """
        Constructs a Bell from a given 0-indexed index.  Should not be used outside this class -
        see `Bell.from_index` and `Bell.from_number` instead.
        """
        if index < 0 or index >= len(BELL_NAMES):
            raise ValueError(f"'{index}' is not known bell index")

        self.index = index

    @property
    def number(self) -> int:
        """ Gets the 1-indexed number representing this bell. """
        return self.index + 1

    def __str__(self) -> str:
        """ Converts this bell to a single-character string representing this bell. """
        return BELL_NAMES[self.index]

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: Any) -> bool:
        """ Determines if two Bells are equal. """
        return isinstance(other, Bell) and other.index == self.index

    def __hash__(self) -> int:
        """ Generates a has of a Bell. """
        return self.index



""" A module to contain the abstract base class that all row generators inherit from. """

import logging
from typing import Dict, List, Tuple, Optional
from abc import ABCMeta, abstractmethod

from wheatley.row_generation.helpers import rounds
from wheatley.aliases import Row, Places
from wheatley.stroke import Stroke, HANDSTROKE
from wheatley.row_generation.helpers import generate_starting_row

class RowGenerator(metaclass=ABCMeta):
    """ Abstract base class for behaviours common to all row generators. """

    logger_name = "ROWGEN"

    def __init__(self, stage: int, start_row: Optional[str] = None) -> None:
        self.stage = stage
        self.custom_start_row = start_row
        self.start_row = generate_starting_row(stage, start_row)
        self.logger = logging.getLogger(self.logger_name)

        self._has_bob = False
        self._has_single = False
        self._index = 0
        self._row = self.start_row

    def reset(self) -> None:
        """ Reset the row generator. """
        self.logger.debug("Reset")

        self._has_bob = False
        self._has_single = False
        self._index = 0
        self._row = self.start_row

    def reset_calls(self) -> None:
        """ Clear the pending call flags. """
        self.logger.debug("Reset calls")
        self._has_bob = False
        self._has_single = False

    def next_row(self, stroke: Stroke) -> Row:
        """
        Generates the next row (with no calls).  `rg.next_row(s)` is exactly equivalent to
        `rg.next_row_and_calls(s)[0]`, and does not call `_gen_row` directly.
        """
        return self.next_row_and_calls(stroke)[0]

    def next_row_and_calls(self, stroke: Stroke) -> Tuple[Row, List[str]]:
        """ Generate the next row, and mutate state accordingly. """
        self._row, calls = self._gen_row_and_calls(self._row, stroke, self._index)

        self._index += 1

        self.logger.debug(self._row)

        return (self._row, calls)

    def set_bob(self) -> None:
        """ Set the flag that a bob has been made. """
        self._has_bob = True

    def set_single(self) -> None:
        """ Set the flag that a single has been made. """
        self._has_single = True

    def rounds(self) -> Row:
        """ Generate rounds of the stage given by this RowGenerator. """
        return rounds(self.stage)

    @abstractmethod
    def _gen_row(self, previous_row: Row, stroke: Stroke, index: int) -> Row:
        pass

    # If this isn't overridden by a child class, then we assume that it won't produce calls, and
    # therefore we simply return no calls on every Row).
    def _gen_row_and_calls(self, previous_row: Row, stroke: Stroke, index: int) -> Tuple[Row, List[str]]:
        return (self._gen_row(previous_row, stroke, index), [])

    def start_stroke(self) -> Stroke: # pylint: disable=no-self-use
        """ Gets the stroke of the first row.  This defaults to HANDSTROKE, but should be overridden by
        other RowGenerators if different start strokes are possible. """
        return HANDSTROKE

    def early_calls(self) -> Dict[int, List[str]]: # pylint: disable=no-self-use
        """
        Returns the calls that should be called **before** going into the changes.  This map
        goes from <number of rows before the first row of method> to <list of calls that should be
        called on that row>.  For example, if we call a 'single' at the start of Original (as in
        https://complib.org/composition/68549), then the `early_calls` should return
        `{2: ["Go Original"], 1: ["Single"]}`.

        If this is not overridden, then it is assumed that the child class doesn't want to call any
        calls, so we return a default empty set
        """
        return {}

    @abstractmethod
    def summary_string(self) -> str:
        """ Returns a short string summarising the RowGenerator.
        This should make grammatical sense when formatted with
        'Wheatley (will ring|is ringing) {summary_string}'.
        """

    def permute(self, row: Row, places: Places) -> Row:
        """ Permute a row by a place notation given by `places`. """
        new_row = list(row)

        i = 1
        if places and places[0] % 2 == 0:
            # Skip 1 for implicit lead when lowest pn is even
            i += 1

        while i < self.stage:
            if i in places:
                i += 1
                continue

            # If not in place, must swap, index is 1 less than place
            new_row[i - 1], new_row[i] = new_row[i], new_row[i - 1]
            i += 2

        return Row(new_row)



""" File to contain string constants representing names used by Ringing Room to represent calls. """

LOOK_TO = "Look to"
GO = "Go"
BOB = "Bob"
SINGLE = "Single"
THATS_ALL = "That's all"
STAND = "Stand next"



import unittest

from wheatley.parsing import parse_peal_speed, PealSpeedParseError, parse_call, CallParseError, \
                             parse_start_row, StartRowParseError, parse_place_notation, PlaceNotationError
from wheatley.aliases import CallDef
from wheatley.row_generation.complib_composition_generator import parse_arg, InvalidComplibURLError


class ParseTests(unittest.TestCase):

    def test_peal_speed_parsing(self):
        test_cases = [("3h04m", 184),
                      ("10", 10),
                      ("3m", 3),
                      ("3h", 180),
                      ("3 h", 180),
                      ("2h58", 178),
                      ("2h58m", 178),
                      (" 2 h 30 m ", 150)]

        for (input_arg, expected_minutes) in test_cases:
            with self.subTest(input=input_arg, expected_minutes=expected_minutes):
                self.assertEqual(expected_minutes, parse_peal_speed(input_arg))

    def test_peal_speed_parsing_errors(self):
        test_cases = [("3h4hhh", "The peal speed should contain at most one 'h'."),
                      ("Xh", "The hour value 'X' is not an integer."),
                      ("-4h", "The hour value '-4' must be a positive integer."),
                      ("3hP", "The minute value 'P' is not an integer."),
                      ("3h-2m", "The minute value '-2' must be a positive integer."),
                      ("3h100", "The minute value '100' must be smaller than 60."),
                      ("125x", "The minute value '125x' is not an integer."),
                      ("-2m", "The minute value '-2' must be a positive integer."),
                      ("-200", "The minute value '-200' must be a positive integer."),
                      ("", "The minute value '' is not an integer."),
                      ("    ", "The minute value '' is not an integer."),
                      ("\nXX   X ", "The minute value 'XX   X' is not an integer.")]

        for (input_arg, expected_message) in test_cases:
            with self.subTest(input=input_arg, expected_message=expected_message):
                with self.assertRaises(PealSpeedParseError) as e:
                    parse_peal_speed(input_arg)
                self.assertEqual(expected_message, e.exception.message)

    def test_call_parsing(self):
        test_cases = [("14", {0: '14'}),
                      ("3.123", {0: '3.123'}),
                      ("  0 \t:  \n 16   ", {0: '16'}),
                      ("20: 70", {20: '70'}),
                      ("20: 70/ 14", {20: '70', 0: '14'})]

        for (input_arg, expected_call_dict) in test_cases:
            with self.subTest(input=input_arg, expected_call_dict=expected_call_dict):
                self.assertEqual(CallDef(expected_call_dict), parse_call(input_arg))

    def test_call_parsing_errors(self):
        test_cases = [("xx:14", "Location 'xx' is not an integer."),
                      ("", "Place notation strings cannot be empty."),
                      ("  /  /    ", "Place notation strings cannot be empty."),
                      ("14/ 1234  ", "Location 0 has two conflicting calls: '14' and '1234'."),
                      (":::  /", "Call specification ':::' should contain at most one ':'.")]

        for (input_arg, expected_message) in test_cases:
            with self.subTest(input=input_arg, expected_message=expected_message):
                with self.assertRaises(CallParseError) as e:
                    parse_call(input_arg)
                self.assertEqual(expected_message, e.exception.message)

    def test_place_notation_parsing(self):
        test_cases = [("5:5.1.5.1.5", (5, "5.1.5.1.5")),
                      ("6:x16,12", (6, "x16,12"))]

        for (input_arg, expected_tuple) in test_cases:
            with self.subTest(input=input_arg, expected_tuple=expected_tuple):
                self.assertEqual(expected_tuple, parse_place_notation(input_arg))

    def test_place_notation_parsing_errors(self):
        test_cases = [("x16x16x16,12", "<stage>:<place notation> required"),
                      ("x:x16x16x16,12", "Stage must be a number"),
                      ("6:z16x16x16,12", "Place notation is invalid")]

        for (input_arg, expected_message) in test_cases:
            with self.subTest(input=input_arg, expected_message=expected_message):
                with self.assertRaises(PlaceNotationError) as e:
                    parse_place_notation(input_arg)
                self.assertEqual(expected_message, e.exception.message)

    def test_url_conversion_public(self):
        # All of these test cases should point to comp #73916, with no access key
        test_cases = [
            "https://complib.org/composition/73916",
            "http://complib.org/composition/73916",
            "http://www.complib.org/composition/73916",
            "complib.org/composition/73916",
            "73916",
            "http://www.complib.org/composition/73916/rows",
            "api.complib.org/composition/73916/rows",
            "http://www.api.complib.org/composition/73916",
            "https://www.api.complib.org/composition/73916/rows",
        ]
        for input_url in test_cases:
            with self.subTest(input=input_url):
                _id, key, substituted_method_id = parse_arg(input_url)
                self.assertEqual(_id, 73916)
                self.assertEqual(key, None)
                self.assertEqual(substituted_method_id, None)

    def test_url_conversion_private(self):
        # All of these test cases should point to comp #65575, with access key "8b9f...da5d"
        test_cases = [
            "https://complib.org/composition/65575?accessKey=8b9fcc4fb35c428c31e711f657a3e4aac1b3da5d",
            "complib.org/composition/65575/rows?accessKey=8b9fcc4fb35c428c31e711f657a3e4aac1b3da5d",
            "www.complib.org/composition/65575?accessKey=8b9fcc4fb35c428c31e711f657a3e4aac1b3da5d",
            "www.complib.org/composition/65575/rows?accessKey=8b9fcc4fb35c428c31e711f657a3e4aac1b3da5d",
            "www.api.complib.org/composition/65575/rows?accessKey=8b9fcc4fb35c428c31e711f657a3e4aac1b3da5d",
            "https://www.api.complib.org/composition/65575/rows?accessKey=8b9fcc4fb35c428c31e711f657a3e4aac1b3da5d",
            "api.complib.org/composition/65575?accessKey=8b9fcc4fb35c428c31e711f657a3e4aac1b3da5d",
            "65575?accessKey=8b9fcc4fb35c428c31e711f657a3e4aac1b3da5d",
        ]
        for input_url in test_cases:
            with self.subTest(input=input_url):
                _id, key, substituted_method_id = parse_arg(input_url)
                self.assertEqual(_id, 65575)
                self.assertEqual(key, "8b9fcc4fb35c428c31e711f657a3e4aac1b3da5d")
                self.assertEqual(substituted_method_id, None)

    def test_url_conversion_subst_method(self):
        test_cases = [
            "https://complib.org/composition/36532?substitutedmethodid=20336",
            "complib.org/composition/36532?substitutedmethodid=20336",
            "www.complib.org/composition/36532?substitutedmethodid=20336",
            "https://www.complib.org/composition/36532?substitutedmethodid=20336",
            "https://www.api.complib.org/composition/36532/rows?substitutedmethodid=20336",
            "api.complib.org/composition/36532/rows?substitutedmethodid=20336",
            "36532?substitutedmethodid=20336",
        ]
        for input_url in test_cases:
            with self.subTest(input=input_url):
                _id, key, substituted_method_id = parse_arg(input_url)
                self.assertEqual(_id, 36532)
                self.assertEqual(key, None)
                self.assertEqual(substituted_method_id, 20336)

    def test_url_conversion_priv_subst_method(self):
        test_cases = [
            "https://complib.org/composition/51155?substitutedmethodid=27600&accessKey=9e1fcd2b11435552cf236be93c7ff73058870995",
            "https://www.complib.org/composition/51155?substitutedmethodid=27600&accessKey=9e1fcd2b11435552cf236be93c7ff73058870995",
            "api.complib.org/composition/51155?substitutedmethodid=27600&accessKey=9e1fcd2b11435552cf236be93c7ff73058870995",
            "complib.org/composition/51155?substitutedmethodid=27600&accessKey=9e1fcd2b11435552cf236be93c7ff73058870995",
            "https://complib.org/composition/51155?accessKey=9e1fcd2b11435552cf236be93c7ff73058870995&substitutedmethodid=27600",
            "https://api.complib.org/composition/51155/rows?accessKey=9e1fcd2b11435552cf236be93c7ff73058870995&substitutedmethodid=27600",
            "https://complib.org/composition/51155/rows?accessKey=9e1fcd2b11435552cf236be93c7ff73058870995&substitutedmethodid=27600",
            "51155?substitutedmethodid=27600&accessKey=9e1fcd2b11435552cf236be93c7ff73058870995"
        ]
        for input_url in test_cases:
            with self.subTest(input=input_url):
                _id, key, substituted_method_id = parse_arg(input_url)
                self.assertEqual(_id, 51155)
                self.assertEqual(key, "9e1fcd2b11435552cf236be93c7ff73058870995")
                self.assertEqual(substituted_method_id, 27600)

    def test_url_conversion_errors(self):
        test_cases = [
            ("complib.org/method/12345", "Not a composition."),
            ("", "Composition ID '' is not a number."),
            ("complib.org", "URL needs more path segments."),
            ("complib.org/composition", "URL needs more path segments."),
            ("complib.org/composition/36532?substitutedmethodid=", "Substituted method ID '' is not a number."),
            ("36532?substitutedmethodid=", "Substituted method ID '' is not a number."),
            ("?substitutedmethodid=pengtekkas", "Composition ID '' is not a number."),
        ]
        for input_url, exp_err_str in test_cases:
            with self.subTest(input=input_url, exp_err_str=exp_err_str):
                with self.assertRaisesRegex(InvalidComplibURLError, exp_err_str):
                    parse_arg(input_url)

    def test_parse_start_row(self):
        test_cases = [("654321", 6),
                      ("1", 1)]

        for (input_arg, expected_output) in test_cases:
            with self.subTest(input=input_arg, expected_output=expected_output):
                output = parse_start_row(input_arg)
                self.assertEqual(output, expected_output)
                
    def test_parse_start_row_errors(self):
        test_cases = [("6", "Start row does not contain bell(s) [1, 2, 3, 4, 5]"),
                      ("64321", "Start row does not contain bell(s) [5]"),
                      ("4321G", "'G' is not known bell symbol"),
                      ("654326", "Start row contains bell 6 mutiple times")]

        for (input_arg, expected_message) in test_cases:
            with self.subTest(input=input_arg, expected_message=expected_message):
                with self.assertRaises(StartRowParseError) as e:
                    parse_start_row(input_arg)
                self.assertEqual(expected_message, e.exception.message)

if __name__ == '__main__':
    unittest.main()
